from abc import ABC, abstractmethod
from typing import TypeVar, Generic, List, Optional

from springdata.domain import Page, Sort, Pageable

T = TypeVar("T")
ID = TypeVar("ID")


class CrudRepository(ABC, Generic[T, ID]):
    """
    Interface for generic CRUD operations on a repository for a specific type.
    """

    @abstractmethod
    async def clear(self) -> None:
        """
        Deletes all entities managed by the repository.
        """
        raise NotImplementedError()

    @abstractmethod
    async def count(self) -> int:
        """
        Returns the number of entities available.

        :return: the number of entities.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete(self, entity: T) -> None:
        """
        Deletes a given entity.

        :param entity: must not be None.
        :raises ValueError: in case the given entity is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_all(self, entities: List[T]) -> None:
        """
        Deletes the given entities.

        :param entities: must not be None. Must not contain None elements.
        :raises ValueError: in case the given entities or one of its entities is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_all_by_id(self, ids: List[ID]) -> None:
        """
        Deletes all entities with the given IDs.

        Entities that aren't found in the persistence store are silently ignored.

        :param ids: must not be None. Must not contain None elements.
        :raises ValueError: in case the given ids or one of its elements is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_by_id(self, id_: ID) -> None:
        """
        Deletes the entity with the given id.

        If the entity is not found in the persistence store it is silently ignored.

        :param id_: must not be None.
        :raises ValueError: if id is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def exists_by_id(self, id_: ID) -> bool:
        """
        Returns whether an entity with the given id exists.

        :param id_: must not be None.
        :return: true if an entity with the given id exists, false otherwise.
        :raises ValueError: if id is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def find_all(self, sort: Optional[Sort] = None) -> List[T]:
        """
        Returns all entities.

        :param sort: the specification to sort the results by, default to None.
        :return: all entities.
        """
        raise NotImplementedError()

    @abstractmethod
    async def find_all_by_id(
        self,
        ids: List[ID],
        sort: Optional[Sort] = None,
    ) -> List[T]:
        """
        Returns all entities with the given IDs.

        If some or all ids are not found, no entities are returned for these IDs.

        :param ids: must not be None nor contain any None values.
        :param sort: the specification to sort the results by, default to None.
        :return: guaranteed to be not None. The size can be equal or less than the number of given ids.
        :raises ValueError: in case the given ids or one of its elements is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def find_by_id(self, id_: ID) -> Optional[T]:
        """
        Retrieves an entity by its id.

        :param id_: must not be None.
        :return: the entity with the given id or None if none found.
        :raises ValueError: if id is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def save(self, entity: T) -> T:
        """
        Saves a given entity.

        Use the returned instance for further operations as the save operation might have changed the entity instance
        completely.

        :param entity: must not be None.
        :return: the saved entity, will never be None.
        :raises ValueError: in case the given entity is None.
        """
        raise NotImplementedError()

    @abstractmethod
    async def save_all(self, entities: List[T]) -> List[T]:
        """
        Saves all given entities.

        :param entities: must not be None nor must it contain None.
        :return: the saved entities; will never be None. The returned iterable will have the same size as the iterable
                 passed as an argument.
        :raises ValueError: in case the given entities or one of its entities is None.
        """
        raise NotImplementedError()


class PagingRepository(Generic[T, ID], CrudRepository[T, ID]):
    """
    Interface to retrieve entities using the pagination.
    """

    @abstractmethod
    async def find_page(
        self,
        pageable: Pageable,
        sort: Optional[Sort] = None,
    ) -> Page[T]:
        """
        Returns a :class:`Page` of entities meeting the paging restriction provided in the :class:`Pageable` object.

        :param pageable: must not be None.
        :param sort: the specification to sort the results by, default to None.
        :return: a page of entities.
        :raises ValueError: in case the :class:`Pageable` is None.
        """
        raise NotImplementedError()



import math
from typing import TypeVar, Generic, List

from springdata.domain.pageable import Pageable

T = TypeVar("T")


class Page(Generic[T]):
    """
    A page is a sublist of a list of objects. It allows gain information about the position of it in the containing
    entire list.
    """

    def __init__(self, content: List[T], pageable: Pageable = None, total: int = None):
        """
        Creates a new :class:`Page`.

        :param content: the content of this page, must not be None.
        :param pageable: the paging information, default to unpaged.
        :param total: the total amount of items available, default to content length. The total might be
                               adapted considering the length of the content given, if it is going to be the content of
                               the last page. This is in place to mitigate inconsistencies.
        """
        if content is None:
            raise ValueError("Content must not be None")
        self.content = content
        self.pageable = pageable or Pageable.unpaged()
        if total is None:
            self.total_elements = len(content)
        elif (
            len(content) > 0
            and self.pageable.is_paged()
            and self.pageable.offset + self.pageable.page_size > total
        ):
            self.total_elements = self.pageable.offset + len(content)
        else:
            self.total_elements = total

    @classmethod
    def empty(cls, pageable: Pageable = None) -> "Page":
        """
        Creates a new empty :class:`Page`.

        :return:
        """
        return Page([], pageable, 0)

    @property
    def number(self) -> int:
        """
        Returns the number of the current :class:`Page`.

        :return: the number of the current :class:`Page`.
        """
        return self.pageable.page_number if self.pageable.is_paged() else 0

    @property
    def number_of_elements(self) -> int:
        """
        Returns the number of elements currently on this :class:`Page`.

        :return: the number of elements currently on this :class:`Page`.
        """
        return len(self.content)

    @property
    def size(self) -> int:
        """
        Returns the size of the :class:`Page`.

        :return: the size of the :class:`Page`.
        """
        return (
            self.pageable.page_size if self.pageable.is_paged() else len(self.content)
        )

    @property
    def total_pages(self) -> int:
        """
        Returns the number of total pages.

        :return: the number of total pages.
        """
        return (
            1
            if self.size == 0 or self.total_elements is None
            else math.ceil(self.total_elements / self.size)
        )

    def has_content(self) -> bool:
        """
        Returns if the :class:`Page` has content at all.

        :return:
        """
        return bool(self.content)

    def has_next(self) -> bool:
        """
        Returns if there is a next :class:`Page`.

        :return: if there is a next :class:`Page`.
        """
        return self.number + 1 < self.total_pages

    def has_previous(self) -> bool:
        """
        Returns if there is a previous :class:`Page`.

        :return: if there is a previous :class:`Page`.
        """
        return self.number > 0

    def is_first(self) -> bool:
        """
        Returns if the current :class:`Page` is the first one.

        :return:
        """
        return not self.has_previous()

    def is_last(self) -> bool:
        """
        Returns if the current :class:`Page` is the last one.

        :return:
        """
        return not self.has_next()

    def next_pageable(self) -> Pageable:
        """
        Returns the :class:`Pageable` to request the next :class:`Page`.

        :return:
        """
        return self.pageable.next() if self.has_next() else Pageable.unpaged()

    def previous_pageable(self) -> Pageable:
        """
        Returns the :class:`Pageable` to request the previous :class:`Page`.

        :return:
        """
        return (
            self.pageable.previous_or_first()
            if self.has_previous()
            else Pageable.unpaged()
        )



from abc import ABC, abstractmethod


class Pageable(ABC):
    """
    Abstract interface for pagination information.
    """

    @property
    @abstractmethod
    def offset(self) -> int:
        """
        Returns the offset to be taken according to the underlying page and page size.

        :return: the offset to be taken.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def page_number(self) -> int:
        """
        Returns the page to be returned.

        :return: the page to be returned
        :raises NotImplementedError: if the object is unpaged.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def page_size(self) -> int:
        """
        Returns the number of items to be returned.

        :return: the number of items of that page
        :raises NotImplementedError: if the object is unpaged.
        """
        raise NotImplementedError()

    @staticmethod
    def of_size(page_size: int) -> "Pageable":
        """
        Creates a new :class:`Pageable` for the first page (page number 0) given page size.

        :param page_size: the size of the page to be returned, must be greater than 0.
        :return: a new :class:`PageRequest`.
        """
        return PageRequest(0, page_size)

    @staticmethod
    def unpaged() -> "Pageable":
        """
        Creates a :class:`Pageable` instance representing no pagination setup.

        :return:
        """
        return Unpaged()

    @abstractmethod
    def first(self) -> "Pageable":
        """
        Return the :class:`Pageable` requesting the first page.

        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def has_previous(self) -> bool:
        """
        Returns if there's a previous :class:`Pageable` we can access from the current one. Will return false in
        case the current one already refers to the first page.

        :return:
        """
        raise NotImplementedError()

    def is_paged(self) -> bool:
        """
        Returns whether the current :class:`Pageable` contains pagination information.

        :return:
        """
        return True

    def is_unpaged(self) -> bool:
        """
        Returns whether the current :class:`Pageable` does not contain pagination information.

        :return:
        """
        return not self.is_paged()

    @abstractmethod
    def next(self) -> "Pageable":
        """
        Returns the :class:`Pageable` requesting the next :class:`Page`.

        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def previous_or_first(self) -> "Pageable":
        """
        Returns the previous :class:`Pageable` or the first :class:`Pageable` if the current one is already the
        first one.

        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def with_page(self, page_number: int) -> "Pageable":
        """
        Creates a new :class:`Pageable` with page number applied.

        :param page_number:
        :return: a new :class:`PageRequest`.
        :raises NotImplementedError: if the object is unpaged and the page number is not zero.
        """
        raise NotImplementedError()


class PageRequest(Pageable):
    """
    Basic implementation of :class:`Pageable`.
    """

    def __init__(self, page: int, size: int):
        """
        Creates a new :class:`PageRequest`. Pages are zero indexed, thus providing 0 for page will return the first
        page.

        :param page: zero-based page index, must not be negative.
        :param size: the size of the page to be returned, must be greater than 0.
        """
        if page < 0:
            raise ValueError("Page index must not be less than zero")
        if size < 1:
            raise ValueError("Page size must not be less than one")
        self._page = page
        self._size = size

    @property
    def offset(self) -> int:
        return self._page * self._size

    @property
    def page_number(self) -> int:
        return self._page

    @property
    def page_size(self) -> int:
        return self._size

    def first(self) -> "PageRequest":
        return PageRequest(0, self._size)

    def has_previous(self) -> bool:
        return self._page > 0

    def next(self) -> "PageRequest":
        return PageRequest(self._page + 1, self._size)

    def previous_or_first(self) -> "PageRequest":
        return self if self._page == 0 else PageRequest(self._page - 1, self._size)

    def with_page(self, page_number: int) -> "PageRequest":
        return PageRequest(page_number, self._size)


class Unpaged(Pageable):
    """
    :class:`Pageable` implementation to represent the absence of pagination information.
    """

    @property
    def offset(self) -> int:
        raise NotImplementedError()

    @property
    def page_number(self) -> int:
        raise NotImplementedError()

    @property
    def page_size(self) -> int:
        raise NotImplementedError()

    def first(self) -> "Pageable":
        return self

    def has_previous(self) -> bool:
        return False

    def is_paged(self) -> bool:
        return False

    def next(self) -> "Pageable":
        return self

    def previous_or_first(self) -> "Pageable":
        return self

    def with_page(self, page_number: int) -> "Pageable":
        if page_number == 0:
            return self
        raise NotImplementedError()



from enum import Enum
from typing import List, ClassVar, Optional


class Direction(Enum):
    """
    Enumeration for sort directions.
    """

    ASC = 1
    DESC = -1

    @staticmethod
    def value_of(name: str) -> "Direction":
        """
        Returns the :class:`Direction` enum with the specified name.

        :param name: the name of the enum constant to be returned.
        :raises ValueError: in case the given value cannot be parsed into an enum value.
        :return the enum constant with the specified name.
        """
        for d in Direction:
            if d.name == name.upper():
                return d
        raise ValueError(
            f"Invalid value {name} for orders given, has to be either 'desc' or 'asc' (case insensitive)"
        )

    @staticmethod
    def values() -> List["Direction"]:
        """
        Returns an array containing the constants of this enum type, in the order they are declared.

        :return an array containing the constants of this enum type, in the order they are declared.
        """
        return [d for d in Direction]

    def is_ascending(self):
        """
        Returns whether the direction is ascending.
        """
        return self == Direction.ASC

    def is_descending(self):
        """
        Returns whether the direction is descending.
        """
        return self == Direction.DESC


class Order:
    """
    PropertyPath implements the pairing of a :class:`Direction` and a property.
    """

    DEFAULT_DIRECTION: ClassVar[Direction] = Direction.ASC

    def __init__(
        self,
        property_: str,
        direction: Optional[Direction] = DEFAULT_DIRECTION,
    ):
        if not property_:
            raise ValueError("Property must not be None or empty")
        self.property = property_
        self.direction = direction or self.DEFAULT_DIRECTION

    @classmethod
    def by(cls, property_: str) -> "Order":
        """
        Creates a new :class:`Order` instance. Takes a single property. :class:`Direction` defaults to
        Order.DEFAULT_DIRECTION.

        :param property_: must not be None or empty.
        """
        return cls(property_)

    @classmethod
    def asc(cls, property_: str) -> "Order":
        """
        Creates a new :class:`Order` instance. Takes a single property. :class:`Direction` is Direction.ASC.

        :param property_: must not be None or empty.
        """
        return cls(property_, Direction.ASC)

    @classmethod
    def desc(cls, property_: str) -> "Order":
        """
        Creates a new :class:`Order` instance. Takes a single property. :class:`Direction` is Direction.DESC.

        :param property_: must not be None or empty.
        """
        return cls(property_, Direction.DESC)

    def is_ascending(self):
        """
        Returns whether sorting for this property shall be ascending.
        """
        return self.direction.is_ascending()

    def is_descending(self):
        """
        Returns whether sorting for this property shall be descending.
        """
        return self.direction.is_descending()

    def with_direction(self, direction: Direction) -> "Order":
        """
        Returns a new :class:`Order` with the given :class:`Direction`.
        """
        return Order(self.property, direction)

    def with_property(self, property_: str) -> "Order":
        """
        Returns a new :class:`Order` with the given property.

        :param property_: must not be None or empty.
        """
        return Order(property_, self.direction)


class Sort:
    """
    Sort option for queries.

    You have to provide at least a list of properties to sort for that must not include None or empty strings. The
    direction defaults to Order.DEFAULT_DIRECTION.
    """

    def __init__(self, orders: List[Order]):
        self.orders = orders

    @classmethod
    def by(
        cls,
        *properties: str,
        direction: Optional[Direction] = Order.DEFAULT_DIRECTION,
    ) -> "Sort":
        """
        Creates a new :class:`Sort` for the given properties.

        :param properties: must not be None or contain None or empty strings.
        :param direction: defaults to Order.DEFAULT_DIRECTION
        :return:
        """
        if not properties:
            raise ValueError("You have to provide at least one property to sort by")
        return cls([Order(p, direction) for p in properties])

    @classmethod
    def unsorted(cls) -> "Sort":
        """
        Returns a :class:`Sort` instance representing no sorting setup at all.

        :return:
        """
        return cls([])

    def ascending(self) -> "Sort":
        """
        Returns a new :class:`Sort` with the current setup but ascending order direction.
        """
        for order in self.orders:
            order.direction = Direction.ASC
        return self

    def descending(self) -> "Sort":
        """
        Returns a new :class:`Sort` with the current setup but descending order direction.
        """
        for order in self.orders:
            order.direction = Direction.DESC
        return self

    def is_sorted(self) -> bool:
        return not bool(self.orders)

    def is_unsorted(self) -> bool:
        return not self.is_sorted()

    def and_(self, sort: "Sort") -> "Sort":
        """
        Returns a new :class:`Sort` consisting of the Orders of the current :class:`Sort` combined with the given ones.

        :param sort: must not be None.
        :return:
        """
        if sort is None:
            raise ValueError("Sort must not be None")
        orders = [order for order in self.orders]
        orders.extend(sort.orders)
        return Sort(orders)

    def get_order_for(self, property_: str) -> Optional[Order]:
        """
        Returns the class:`Order` registered for the given property.

        :param property_:
        :return:
        """
        for order in self.orders:
            if order.property == property_:
                return order
        return None

    def with_direction(self, direction: Direction) -> "Sort":
        """
        Creates a new :class:`Sort` with the current setup but the given order direction.

        :param direction:
        :return:
        """
        return Sort([Order(order.property, direction) for order in self.orders])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['springdata', 'springdata.asyncio', 'springdata.domain']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'spring-data-python',
    'version': '0.1.6',
    'description': 'Spring Data Python is an offshoot of the Java-based Spring Data Framework, targeted for Python.',
    'long_description': '# spring-data-python\nSpring Data Python is an offshoot of the Java-based Spring Data Framework, targeted for Python. \n',
    'author': 'Vincent TERESE',
    'author_email': 'vincent.terese@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kobibleu/spring-data-python',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
